/* Copyright (c) 2008 Boulder Real Time Technologies, Inc. */
/* All rights reserved */
 
/* This software module is wholly owned by Boulder Real Time 
   Technologies, Inc. This software may be used freely in any 
   way as long as the copyright statement above is not removed. */

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#define __USE_GNU
#include <dlfcn.h>

#include "stack.h"
#include "wffil.h"

#define SAMP(x,y)       (int)((x)<0.0?((x)*(y)-0.5):((x)*(y)+0.5))
#define INT(x)  	(int)((x)<0.0?((x)-0.5):((x)+0.5))

typedef struct stackthread_ {
	int bufsamp_n;
	int bufsamp_n2;
	double dt;
	int ret;
	Stack *st;
} StackThread;

static int stack_process_buffer (Stack *st, int bufsamp_n, int bufsamp_n2, double dt);
static long stack_comp_sub (long ioffset, long number, void *user_data);
static int stack_process_slice (Stack *st, int node_offset, int node_n, 
			double dt, int bufsamp_n, int bufsamp_n2);

static int
stack_process_buffer (Stack *st, int bufsamp_n, int bufsamp_n2, double dt)

{
	int i;

	/* setup and launch processing threads */

	if (st->threads_tbl == NULL) {
		int node_offset, node_n;

		if (st->threads_n < 0) st->threads_n = 1;
		if (st->threads_n > st->nnodes) st->threads_n = st->nnodes;
		st->threads_tbl = inittbl(0,st->threads_n,1,NULL,sizeof(StackThread));
		if (st->threads_tbl == NULL) {
			register_error (0, "stack_process_buffer: inittbl(threads_tbl) error.\n");
			return (-1);
		}
		st->threads = pmtcomp_create();
		if (st->threads == NULL) {
			register_error (0, "stack_process_buffer: pmtcomp_create(threads) error.\n");
			return (-1);
		}
		node_offset = 0;
		node_n = st->nnodes / st->threads_n;
		if (node_n < 1) node_n = 1;
		for (i=0; i<st->threads_n; i++) {
			if (node_offset + node_n > st->nnodes) node_n = st->nnodes - node_offset;
			if (i == st->threads_n-1) node_n = st->nnodes - node_offset;
			if (node_n < 1) {
				st->threads_n = i;
				break;
			}
			node_offset += node_n;
		}
		node_offset = 0;
		node_n = st->nnodes / st->threads_n;
		if (node_n < 1) node_n = 1;
		for (i=0; i<st->threads_n; i++) {
			StackThread thrv, *thr;
	
			memset (&thrv, 0, sizeof(StackThread));
			settbl (st->threads_tbl, i, &thrv);
			thr = (StackThread *) gettbl (st->threads_tbl, i);
			if (node_offset + node_n > st->nnodes) node_n = st->nnodes - node_offset;
			if (i == st->threads_n-1) node_n = st->nnodes - node_offset;
			if (st->threads_n > 1) {
				if (pmtcomp_addthread (st->threads, stack_comp_sub, node_offset, node_n, thr) < 0) {
					register_error (0, "stack_process_buffer: pmtcomp_addthread(StackThread %d) error.\n", i);
					return (-1);
				}
			}
			node_offset += node_n;
		}
	}

	/* setup thread parameters*/

	for (i=0; i<st->threads_n; i++) {
		StackThread *thr;

		thr = (StackThread *) gettbl (st->threads_tbl, i);
		thr->bufsamp_n = bufsamp_n;
		thr->bufsamp_n2 = bufsamp_n2;
		thr->dt = dt;
		thr->st = st;
	}

	/* process threads */

	if (st->threads_n > 1) {
		if (pmtcomp_processthreads (st->threads) != st->threads_n) {
			register_error (0, "stack_process_buffer: pmtcomp_processthreads() error.\n");
			return (-1);
		}
	} else {
		StackThread *thr;

		thr = (StackThread *) gettbl (st->threads_tbl, 0);
		if (stack_comp_sub (0, st->nnodes, thr) < 0) {
			register_error (0, "stack_process_buffer: stack_comp_sub() error.\n");
			return (-1);
		}
	}

	return (0);
}

static long 
stack_comp_sub (long ioffset, long number, void *user_data)

{
	StackThread *thr = (StackThread *) user_data;

	return (stack_process_slice (thr->st, ioffset, number, thr->dt, thr->bufsamp_n, thr->bufsamp_n2));
}

static int
stack_process_slice (Stack *st, int node_offset, int node_n, double dt, int bufsamp_n, int bufsamp_n2)

{
	int inode, bufsamp;
	size_t size;

	for (inode=node_offset; inode<node_offset+node_n; inode++) {
		StackNode *node;
		int nsamps;
		double dt2, tstart;
		float *dptr;

		node = (StackNode *) gettbl (st->nodes_tbl, inode);

		if (node->stack == NULL) {
			size = bufsamp_n*sizeof(float);
			node->stack = (float *) malloc (size);
			if (node->stack == NULL) {
				register_error (1, "stack_process_slice: malloc(node->stack,%ld) error.\n", size);
				return (-1);
			}
			memset (node->stack, 0, size);
			node->stackpower = (float *) malloc (size);
			if (node->stackpower == NULL) {
				register_error (1, "stack_process_slice: malloc(node->stackpower,%ld) error.\n", size);
				free (node->stack);
				node->stack = NULL;
				return (-1);
			}
			memset (node->stackpower, 0, size);
			node->powerstack = (float *) malloc (size);
			if (node->powerstack == NULL) {
				register_error (1, "stack_process_slice: malloc(node->powerstack,%ld) error.\n", size);
				free (node->stack);
				free (node->stackpower);
				node->stack = NULL;
				return (-1);
			}
			memset (node->powerstack, 0, size);
			node->nsamps = bufsamp_n;
		} else if (bufsamp_n > node->nsamps) {
			size = bufsamp_n*sizeof(float);
			node->stack = (float *) realloc (node->stack, size);
			if (node->stack == NULL) {
				free (node->stackpower);
				free (node->powerstack);
				register_error (1, "stack_process_slice: realloc(node->stack,%ld) error.\n", size);
				return (-1);
			}
			memset (node->stack, 0, size);
			node->stackpower = (float *) realloc (node->stackpower, size);
			if (node->stackpower == NULL) {
				register_error (1, "stack_process_slice: realloc(node->stackpower,%ld) error.\n", size);
				free (node->stack);
				free (node->powerstack);
				node->stack = NULL;
				return (-1);
			}
			memset (node->stackpower, 0, size);
			node->powerstack = (float *) realloc (node->powerstack, size);
			if (node->powerstack == NULL) {
				register_error (1, "stack_process_slice: realloc(node->powerstack,%ld) error.\n", size);
				free (node->stack);
				free (node->stackpower);
				node->stack = NULL;
				return (-1);
			}
			memset (node->powerstack, 0, size);
			node->nsamps = bufsamp_n;
		}

		for (bufsamp=0; bufsamp<bufsamp_n; bufsamp++) {
			float stack, powerstack;
			int nstack, nstackp, j;

			stack = st->gap_value;
			powerstack = st->gap_value;
			nstack = 0;
			nstackp = 0;
			for (j=0; j<st->nchannels; j++) {
				int shiftsamp;
				float value, powervalue;

				shiftsamp = bufsamp+node->shifts[j]-st->shift_min;
				value = st->gap_value;
				powervalue = st->gap_value;
				if (shiftsamp < 0) {
					complain (0, "stack_process_slice: shiftsamp < 0 - should never happen\n");
				} else if (shiftsamp >= bufsamp_n2) {
					complain (0, "stack_process_slice: shiftsamp >= n - should never happen\n");
				} else {
					value = st->gatherbuffer[shiftsamp+j*st->nsampsbuf];
					powervalue = st->gatherpowerbuffer[shiftsamp+j*st->nsampsbuf];
					if (powervalue < 0.0) {
						complain (0, "stack_process_slice: powervalue < 0 (%.4f) - should never happen\n", powervalue);
					} 
					if (powervalue != st->gap_value && powervalue > 1.e30) {
						complain (0, "stack_process_slice: wrong looking powervalue (%.4f) at %d %d\n", powervalue, j, bufsamp);
					}
				}
				if (value == st->gap_value) continue;
				if (value == st->clip_value) continue;
				if (stack == st->gap_value) {
					stack = value;
					nstack = 1;
				} else {
					stack += value;
					nstack++;
				}
				if (powervalue == st->gap_value) continue;
				if (powerstack == st->gap_value) {
					powerstack = powervalue;
					nstackp = 1;
				} else {
					powerstack += powervalue;
					nstackp++;
				}
			}
			if (nstack < 1 || nstack < st->nchannels-st->nallowedbadchannels) {
				stack = st->gap_value;
			} else {
				if (stack != st->gap_value) stack /= nstack;
			}
			if (nstackp < 1 || nstackp < st->nchannels-st->nallowedbadchannels) {
				powerstack = st->gap_value;
			} else {
				if (powerstack != st->gap_value) powerstack /= nstackp;
			}
			node->stack[bufsamp] = stack;
			node->powerstack[bufsamp] = powerstack;
			if (node->powerstack[bufsamp] != st->gap_value && node->powerstack[bufsamp] > 1.e30) {
				complain (0, "stack_process_slice: wrong looking powerstack (%.4f) at %d %d\n", node->powerstack[bufsamp], inode, bufsamp);
			}
		}
	
		size = bufsamp_n*sizeof(float);
		memcpy (node->stackpower, node->stack, size);
		dptr = node->stackpower;
		nsamps = bufsamp_n;
		dt2 = dt;
		tstart = st->buftime;
		if (wffil (NULL, st->power_filter, (double) 0.5, &nsamps,
				&tstart, &dt2, &dptr, &nsamps, 
				NULL, NULL, &(node->stack_power_filter)) < 0) {
			register_error (0, "stack_process_slice: wffil(stack) error.\n");
			return (-1);
		}
	}

	return (0);
}

static Arr *moveout_class_arr=NULL;

MoveoutClass *
moveout_class_get (char *class_name, int verbose)

{
	MoveoutClass *class=NULL;
	char string[128];
	void * (*_params_new) (Pf *params);
	void   (*_params_destroy) (void *params);
	void * (*_params_copy) (void *params);
	int   (*_nodes_new) (Stack *st);
	void   (*_nodes_destroy) (Stack *st);
	char * (*_nodes_print) (void *params, int verbose);
	int   (*_data_topkt_check) (void *params, char *channel, PktChannel *chan);
	int   (*_data_topkt_data) (void *params, char *channel, int inode, int *data);
	int   (*_data_nodes_togrid) (void *params, BUGrid **grid);
	size_t size;

	if (moveout_class_arr == NULL) {
		moveout_class_arr = newarr (0);
		if (moveout_class_arr == NULL) {
			die (0, "newarr(moveout_class_arr) error.\n");
		}
	}

	class = getarr (moveout_class_arr, class_name);
	if (class) return (class);

	if (strlen(class_name) > 3 
			&& !strcmp(&(class_name[strlen(class_name)-3]), ".pm")) {
		register_error (0, "moveout_class_get: perl classes not supported.\n");
		return (NULL);
	}

	sprintf (string, "%s_params_new", class_name);
	*(void **)&_params_new = dlsym (RTLD_DEFAULT, string);

	if (_params_new != NULL) {
		if (verbose) {
			elog_notify (0, "Loading stack class '%s' from local address space.\n", class_name);
		}

		sprintf (string, "%s_params_destroy", class_name);
		*(void **)&_params_destroy = dlsym (RTLD_DEFAULT, string);
		if (_params_destroy == NULL) {
			register_error (0, "moveout_class_get: cannot find %s procedure in self.\n", string);
			return (NULL);
		}
		sprintf (string, "%s_params_copy", class_name);
		*(void **)&_params_copy = dlsym (RTLD_DEFAULT, string);
		if (_params_copy == NULL) {
			register_error (0, "moveout_class_get: cannot find %s procedure in self.\n", string);
			return (NULL);
		}
		sprintf (string, "%s_nodes_new", class_name);
		*(void **)&_nodes_new = dlsym (RTLD_DEFAULT, string);
		if (_nodes_new == NULL) {
			register_error (0, "moveout_class_get: cannot find %s procedure in self.\n", string);
			return (NULL);
		}
		sprintf (string, "%s_nodes_destroy", class_name);
		*(void **)&_nodes_destroy = dlsym (RTLD_DEFAULT, string);
		if (_nodes_destroy == NULL) {
			register_error (0, "moveout_class_get: cannot find %s procedure in self.\n", string);
			return (NULL);
		}
		sprintf (string, "%s_nodes_print", class_name);
		*(void **)&_nodes_print = dlsym (RTLD_DEFAULT, string);
		if (_nodes_print == NULL) {
			register_error (0, "moveout_class_get: cannot find %s procedure in self.\n", string);
			return (NULL);
		}
		sprintf (string, "%s_data_topkt_check", class_name);
		*(void **)&_data_topkt_check = dlsym (RTLD_DEFAULT, string);
		if (_data_topkt_check == NULL) {
			register_error (0, "moveout_class_get: cannot find %s procedure in self.\n", string);
			return (NULL);
		}
		sprintf (string, "%s_data_topkt_data", class_name);
		*(void **)&_data_topkt_data = dlsym (RTLD_DEFAULT, string);
		if (_data_topkt_data == NULL) {
			register_error (0, "moveout_class_get: cannot find %s procedure in self.\n", string);
			return (NULL);
		}
		sprintf (string, "%s_data_nodes_togrid", class_name);
		*(void **)&_data_nodes_togrid = dlsym (RTLD_DEFAULT, string);
		if (_data_nodes_togrid == NULL) {
			register_error (0, "moveout_class_get: cannot find %s procedure in self.\n", string);
			return (NULL);
		}
	} else {
		register_error (0, "moveout_class_get: dlopen classes not supported.\n");
		return (NULL);
	}

	size = sizeof(MoveoutClass);
	class = (MoveoutClass *) malloc (size);
	if (class == NULL) {
		die (1, "malloc(MoveoutClass,%ld) error.\n", size);
	}

	memset (class, 0, size);

	strcpy (class->name, class_name);
	class->type = WFPROC_CLASS_TYPE_COMPILED;
	class->_params_new 		= _params_new;
	class->_params_destroy		= _params_destroy;
	class->_params_copy		= _params_copy;
	class->_nodes_new 		= _nodes_new;
	class->_nodes_destroy		= _nodes_destroy;
	class->_nodes_print		= _nodes_print;
	class->_data_topkt_check	= _data_topkt_check;
	class->_data_topkt_data		= _data_topkt_data;
	class->_data_nodes_togrid	= _data_nodes_togrid;

	setarr (moveout_class_arr, class_name, class);

	return (class);
}

Stack *
stack_new (Pf *params, int verbose)

{
	char strbuf[512];
	int ret;
	size_t size;
	WFProcDb *wfpdbm;
	Dbptr dbmaster;
	double time;
	Stack *st;
	Dbptr site;
	char *ptr;
	char snet[32];
	char ssta[32];
	char expr[512];
	int i, yd;
	long n;
	Tbl *outputs=NULL;

	if (parse_param (params, "dbmaster", P_STRCP, 1, strbuf) < 0) {
		register_error (0, "stack_new: parse_param(dbmaster) error.\n");
		return (NULL);
	}

	ret = sscanf (strbuf, "%lu", (unsigned long *) &wfpdbm);
	if (ret != 1) {
		register_error (0, "stack_new: Cannot parse 'dbmaster' parameter (%s).\n", strbuf);
		return (NULL);
	}

	if (wfprocdb_lock (wfpdbm, &dbmaster) < 0) {
		register_error (0, "stack_new: wfprocdb_lock() error.\n");
		return (NULL);
	}

	strcpy (strbuf, "");
	if (parse_param (params, "time", P_STRCP, 0, strbuf) < 0) {
		wfprocdb_unlock(wfpdbm);
		register_error (0, "stack_new: parse_param(time) error.\n");
		return (NULL);
	}
	if (!strcmp(strbuf, "")) {
		time = now();
	} else {
		if (zstr2epoch (strbuf, &time) < 0) {
			wfprocdb_unlock(wfpdbm);
			register_error (0, "stack_new: Cannot parse time value '%s'.\n", strbuf);
			return (NULL);
		}
	}

	size = sizeof(Stack);
	st = (Stack *) malloc (size);
	if (st == NULL) {
		wfprocdb_unlock(wfpdbm);
		register_error (1, "stack_new: malloc(Stack,%ld) error.\n", size);
		return (NULL);
	}

	memset (st, 0, size);

	st->wfpdbm = wfpdbm;
	st->dbmaster = dbmaster;
	st->time = time;
	st->latr = -999.0;
	st->lonr = -999.0;

	if (parse_param (params, "type", P_STRCP, 1, st->moveout_class_name) < 0) {
		wfprocdb_unlock(wfpdbm);
		register_error (0, "stack_new: parse_param(type) error.\n");
		return (NULL);
	}
	st->moveout_class = moveout_class_get (st->moveout_class_name, verbose);
	if (st->moveout_class == NULL) {
		wfprocdb_unlock(wfpdbm);
		register_error (0, "stack_new: moveout_class_get(%s) error.\n", st->moveout_class_name);
		return (NULL);
	}
	st->moveout_params = (*(st->moveout_class->_params_new)) (params);
	if (st->moveout_params == NULL) {
		wfprocdb_unlock(wfpdbm);
		register_error (0, "stack_new: %s_params_new() error.\n", st->moveout_class_name);
		return (NULL);
	}

	if (parse_param (params, "refsta", P_STRCP, 1, st->refsta_seed) < 0) {
		wfprocdb_unlock(wfpdbm);
		register_error (0, "stack_new: parse_param(refsta) error.\n");
		return (NULL);
	}
	ptr = strchr (st->refsta_seed, '_');
	if (ptr == NULL) {
		strcpy (st->refsta, st->refsta_seed);
		if (dbseed_net (dbmaster, st->refsta, snet, ssta) < 0) {
			wfprocdb_unlock(wfpdbm);
			register_error (0, "stack_new: seed_net(refsta=%s) error.\n",
							st->refsta_seed);
			return (NULL);
		}
		sprintf (st->refsta_seed, "%s_%s", snet, ssta);
	} else {
		*ptr = '\0';
		ptr++;
		strcpy (snet, st->refsta_seed);
		strcpy (ssta, ptr);
		ptr--;
		*ptr = '_';
		dbmap_seed_netsta (dbmaster, snet, ssta, st->refsta);
		clear_register (1);
	}

	site = dblookup (st->dbmaster, 0, "site", 0, 0);
	if (site.table == dbINVALID) {
		register_error (0, "stack_new: Cannot lookup site table.\n");
		return (NULL);
	}
	yd = yearday(st->time);
	sprintf (expr, "%d >= ondate && ( %d <= offdate || offdate == null(\"offdate\") ) && sta == \"%s\"", 
					yd, yd, st->refsta);
	site = dbsubset (site, expr, NULL);
	dbquery (site, dbRECORD_COUNT, &n);
	if (n < 1) {
		wfprocdb_unlock(wfpdbm);
		register_error (0, "stack_new: Cannot find refsta '%s' in site table for yearday %d.\n",
							st->refsta, yd);
		return (NULL);
	}
	if (n > 1) {
		wfprocdb_unlock(wfpdbm);
		register_error (0, "stack_new: refsta '%s' appears more than once in site table for yearday %d.\n",
							st->refsta, yd);
		return (NULL);
	}
	site.record = 0;
	if (dbgetv (site, NULL, "lat", &(st->latr), "lon", &(st->lonr), NULL) < 0) {
		wfprocdb_unlock(wfpdbm);
		register_error (0, "stack_new: dbgetv(lat,lon) error for refsta '%s'.\n",
							st->refsta);
		return (NULL);
	}
	wfprocdb_unlock(wfpdbm);

	if (parse_param (params, "number_allowed_bad_channels", P_LINT, 0, &(st->nallowedbadchannels)) < 0) {
		register_error (0, "stack_new: parse_param(number_allowed_bad_channels) error.\n");
		return (NULL);
	}
	if (parse_param (params, "twin_power", P_DBL, 1, &(st->twin_power)) < 0) {
		register_error (0, "stack_new: parse_param(twin_power) error.\n");
		return (NULL);
	}
	if (st->twin_power > 0.0) {
		sprintf (st->power_filter, "SQ;AVE %.3f 0.0 100.0", st->twin_power);
	} else {
		strcpy (st->power_filter, "none");
	}
	if (parse_param (params, "grid_decimation", P_LINT, 0, &(st->grid_decimation)) < 0) {
		register_error (0, "stack_new: parse_param(grid_decimation) error.\n");
		return (NULL);
	}
	if (st->grid_decimation < 1) st->grid_decimation = 1;

	if (parse_param (params, "outputs", P_TBL, 1, &outputs) < 0) {
		register_error (0, "stack_new: parse_param(outputs) error.\n");
		return (NULL);
	}
	st->outputs = newarr(NULL);
	if (st->outputs == NULL) {
		register_error (0, "stack_new: newarr(outputs) error.\n");
		return (NULL);
	}

	for (i=0; i<maxtbl(outputs); i++) {
		char *ptr;
		char line[512];
		char output_variable[64];
		char output_type[8];
		char seed_name[128];
		char nodes[128];
		int ret;
		size_t size;
		Tbl *pktoutput_channels;
		StackPktoutputType *pktoutput_channel;

		ptr = (char *) gettbl (outputs, i);
		strcpy (line, ptr);
		ptr = strchr(line, '#');
		if (ptr) *ptr = '\0';
		strtrim(line);
		if (strlen(line) < 1) continue;

		ret = sscanf (line, "%s %s %s %s", output_variable, output_type, seed_name, nodes);
		if (ret < 3 || ret > 4) {
			register_error (0, "stack_new: Cannot parse outputs line %s.\n", line);
			return (NULL);
		}

		if (!strcmp(output_variable, "stack")) {
		} else if (!strcmp(output_variable, "stackpower")) {
		} else if (!strcmp(output_variable, "powerstack")) {
		} else if (!strcmp(output_variable, "semblance")) {
		} else {
			register_error (1, "stack_new: Unrecognized output_variable '%s'.\n", output_variable);
			return (NULL);
		}

		if (!strcmp(output_type, "wf")) {
			if (ret != 4) {
				register_error (0, "stack_new: Cannot parse outputs line %s.\n", line);
				return (NULL);
			}
		} else if (!strcmp(output_type, "grid")) {
		} else {
			register_error (1, "stack_new: Unrecognized output_type '%s'.\n", output_type);
			return (NULL);
		}

		pktoutput_channels = (Tbl *) getarr (st->outputs, output_variable);
		if (pktoutput_channels == NULL) {
			pktoutput_channels = newtbl(0);
			if (pktoutput_channels == NULL) {
				register_error (0, "stack_new: newtbl(outputs->%s) error.\n", output_variable);
				return (NULL);
			}
			setarr (st->outputs, output_variable, pktoutput_channels);
		}

		size = sizeof(StackPktoutputType);
		pktoutput_channel = (StackPktoutputType *) malloc (size);
		if (pktoutput_channel == NULL) {
			register_error (1, "stack_new: malloc(StackPktoutputType,%ld) error.\n", size);
			return (NULL);
		}
		memset (pktoutput_channel, 0, size);
		strcpy (pktoutput_channel->output_type, output_type);
		strcpy (pktoutput_channel->seed_name, seed_name);
		strcpy (pktoutput_channel->nodes, nodes);
		settbl (pktoutput_channels, -1, pktoutput_channel);
	}
	freetbl (outputs, NULL);

	st->threads_n = 1;
	if (parse_param (params, "number_threads", P_LINT, 0, &(st->threads_n)) < 0) {
		register_error (0, "stack_new: parse_param(threads_n) error.\n");
		return (NULL);
	}
	if (st->threads_n < 1) st->threads_n = 1;
	
	return (st);
}

int 
stack_destroy (Stack *st)

{
	if (st == NULL) return (0);

	if (st->threads) pmtcomp_destroy (st->threads);

	if (st->threads_tbl) freetbl (st->threads_tbl, NULL);
	if (st->channels_tbl) freetbl (st->channels_tbl, free);
	if (st->nodes_tbl) freetbl (st->nodes_tbl, free);
	if (st->gatherbuffer) free (st->gatherbuffer);
	if (st->gatherpowerbuffer) free (st->gatherpowerbuffer);

	free (st);

	return (0);
}

static void stack_mybugridfree (void *ptr);

int
stack_proc (Stack *st, WFProcData *input, Tbl **output, Pf **state)

{
	WFGatherOutput *data;
	WFGatherOutput *powerdata;
	WFProcData *powerinput;
	WFProcData *out;
	size_t size;
	int samp, samp0, bufsamp, j, n, n2;
	double dt;
	Tbl *outputs;

	*output = newtbl(0);
	if (*output == NULL) {
		register_error (0, "stack_proc: newtbl(output) error.\n");
		wfproc_data_free (input);
		return (-1);
	}

	if (strcmp(input->type, "wfgather")) {
		wfproc_data_free (input);
		return (0);
	}

	data = (WFGatherOutput *) input->data;

	if (st->channels_tbl == NULL) {
		st->channels_tbl = wfgather_channels_copy (data->channels);
		if (st->channels_tbl == NULL) {
			register_error (0, "stack_proc: wfgather_channels_copy() error.\n");
			wfproc_data_free (input);
			return (-1);
		}
		st->nchannels = maxtbl(st->channels_tbl);
		for (j=0; j<st->nchannels; j++) {
			WFGatherChannel *chan;

			chan = (WFGatherChannel *) gettbl (st->channels_tbl, j);
			chan->filter_hook = NULL;
		}
		st->samprate = data->samprate;
		wfproc_mutex_lock ("stackclass");
		if ((*(st->moveout_class->_nodes_new)) (st) < 0) {
			wfproc_mutex_unlock ("stackclass");
			register_error (0, "stack_proc: %s_nodes_new() error.\n", st->moveout_class_name);
			wfproc_data_free (input);
			return (-1);
		}
		wfproc_mutex_unlock ("stackclass");
		j = st->shift_max-st->shift_min;
		n = (j/data->nsamps) + 2;
		st->nsampsbuf = n*data->nsamps;
		st->buftime_shift = st->shift_min/st->samprate;
		st->buftime = data->tstart + st->buftime_shift;
		size = st->nchannels*st->nsampsbuf*sizeof(float);
		st->gatherbuffer = (float *) malloc (size);
		if (st->gatherbuffer == NULL) {
			register_error (1, "stack_proc: malloc(gatherbuffer,%ld) error.\n", size);
			wfproc_data_free (input);
			return (-1);
		}
		for (samp=0; samp<st->nchannels*st->nsampsbuf; samp++) 
						st->gatherbuffer[samp] = data->gap_value;
		st->gatherpowerbuffer = (float *) malloc (size);
		if (st->gatherpowerbuffer == NULL) {
			register_error (1, "stack_proc: malloc(gatherpowerbuffer,%ld) error.\n", size);
			wfproc_data_free (input);
			return (-1);
		}
		for (samp=0; samp<st->nchannels*st->nsampsbuf; samp++) 
						st->gatherpowerbuffer[samp] = data->gap_value;
	} else {
		if (!wfgather_channels_match (data->channels, st->channels_tbl)) {
			wfproc_data_free (input);
			return (0);
		}
	}

	/* Compute power of the input channel data */

	powerinput = wfproc_data_copy (input);
	if (powerinput == NULL) {
		register_error (0, "stack_proc: wfproc_data_copy(powerinput) error.\n");
		wfproc_data_free (input);
		return (-1);
	}
	powerdata = (WFGatherOutput *) powerinput->data;

	dt = 1.0/st->samprate;

	for (j=0; j<st->nchannels; j++) {
		int nsamps;
		double tstart, dt2;
		WFGatherChannel *chan;
		float *dptr;

		chan = (WFGatherChannel *) gettbl (st->channels_tbl, j);
		nsamps = powerdata->nsamps;
		dt2 = dt;
		tstart = powerdata->tstart;
		dptr = &(powerdata->gather[j*powerdata->nsamps]);
		if (wffil (NULL, st->power_filter, (double) 0.5, &nsamps,
				&tstart, &dt2, &dptr, &nsamps, 
				NULL, NULL, &(chan->filter_hook)) < 0) {
			register_error (0, "stack_proc: wffil(powerdata) error.\n");
			wfproc_data_free (powerinput);
			wfproc_data_free (input);
			return (-1);
		}
		for (samp=0; samp<powerdata->nsamps; samp++) {
			if (powerdata->gather[samp+j*powerdata->nsamps] < 0.0) {
				complain (0, "stack_proc: powervalue < 0 (%.4f) - should never happen\n", 
						powerdata->gather[samp+j*powerdata->nsamps]);
			} 
		}
	}

	/* Figure where the input data shifts into the gather buffers */

	bufsamp = SAMP(data->tstart-st->buftime,st->samprate);
	if (bufsamp < 0) {
		register_error (0, "stack_proc: buffer sample index < 0\n");
		wfproc_data_free (input);
		wfproc_data_free (powerinput);
		return (0);
	}

	/* Resize/shift gather buffers and initialize new values with gap sample values */

	if (bufsamp+data->nsamps > st->nsampsbuf) {
		float *gatherbuffer;
		float *gatherpowerbuffer;

		n = bufsamp+data->nsamps;
		size = n*st->nchannels*sizeof(float);
		gatherbuffer = (float *) malloc (size);
		if (gatherbuffer == NULL) {
			register_error (1, "stack_proc: malloc(gatherbuffer,%ld) error.\n", size);
			wfproc_data_free (input);
			wfproc_data_free (powerinput);
			return (-1);
		}
		gatherpowerbuffer = (float *) malloc (size);
		if (gatherpowerbuffer == NULL) {
			register_error (1, "stack_proc: malloc(gatherpowerbuffer,%ld) error.\n", size);
			wfproc_data_free (input);
			wfproc_data_free (powerinput);
			return (-1);
		}
		for (samp=0; samp<n*st->nsampsbuf; samp++) {
			gatherbuffer[samp] = data->gap_value;
			gatherpowerbuffer[samp] = data->gap_value;
		}
		for (samp=0; samp<st->nsampsbuf; samp++) {
			for (j=0; j<st->nchannels; j++) {
				gatherbuffer[samp+j*n] = st->gatherbuffer[samp+j*st->nsampsbuf];
				gatherpowerbuffer[samp+j*n] = st->gatherpowerbuffer[samp+j*st->nsampsbuf];
			}
		}
		st->nsampsbuf = n;
		free (st->gatherbuffer);
		st->gatherbuffer = gatherbuffer;
		free (st->gatherpowerbuffer);
		st->gatherpowerbuffer = gatherpowerbuffer;
	}

	/* Copy input data and power of input data into gather buffers */

	for (samp=0,samp0=bufsamp; samp<data->nsamps; samp++) {
		for (j=0; j<st->nchannels; j++) {
			st->gatherbuffer[bufsamp+j*st->nsampsbuf] 
					= data->gather[samp+j*data->nsamps];
			st->gatherpowerbuffer[bufsamp+j*st->nsampsbuf] 
					= powerdata->gather[samp+j*data->nsamps];
			if (st->gatherpowerbuffer[bufsamp+j*st->nsampsbuf] < 0.0) {
				complain (0, "stack_proc: gatherpowerbuffer < 0 (%.4f) - should never happen\n", 
						st->gatherpowerbuffer[bufsamp+j*st->nsampsbuf]);
			}
		}
		bufsamp++;
	}
	n2 = bufsamp;
	n = bufsamp - (st->shift_max-st->shift_min);
	wfproc_data_free (powerinput);

	if (n < 1) {
		wfproc_data_free (input);
		return (0);
	}

	/* Do the stacking */

	st->gap_value = data->gap_value;
	st->clip_value = data->clip_value;
	if (stack_process_buffer (st, n, n2, dt) < 0) {
		register_error (0, "stack_proc: stack_process_buffer() error.\n");
		wfproc_data_free (input);
		return (-1);
	}

	/* Populate output packets */

	outputs = keysarr (st->outputs);
	for (j=0; j<maxtbl(outputs); j++) {
		char *output_variable;
		Tbl *pktoutput_channels;

		output_variable = (char *) gettbl (outputs, j);
		pktoutput_channels = (Tbl *) getarr (st->outputs, output_variable);
		out = stack_create_output (st, n, output_variable, pktoutput_channels, data);
		if (out == NULL) {
			complain (0, "stack_proc: stack_create_output(%s) error.\n", output_variable);
		} else {
			settbl (*output, -1, out);
		}
	}
	freetbl (outputs, NULL);

	/* increment the gather buffer start time and shift the data values */

	st->buftime += n*dt;
	for (samp=0,bufsamp=n; bufsamp<st->nsampsbuf; bufsamp++,samp++) {
		for (j=0; j<st->nchannels; j++) {
			st->gatherbuffer[samp+j*st->nsampsbuf] =
					st->gatherbuffer[bufsamp+j*st->nsampsbuf];
			st->gatherpowerbuffer[samp+j*st->nsampsbuf] =
					st->gatherpowerbuffer[bufsamp+j*st->nsampsbuf];
		}
	}
	for (; samp<st->nsampsbuf; samp++) {
		for (j=0; j<st->nchannels; j++) {
			st->gatherbuffer[samp+j*st->nsampsbuf] = data->gap_value;
			st->gatherpowerbuffer[samp+j*st->nsampsbuf] = data->gap_value;
		}
	}

	wfproc_data_free (input);

	return (0);
}

WFProcData *
stack_create_output (Stack *st, int nsamps, char *output_variable, Tbl *pktoutput_channels, WFGatherOutput *gdata)

{
	WFProcData *data=NULL;
	StackOutput *out;
	int i, j, k;
	size_t size, size_out;

	size = sizeof(StackOutput);
	size_out = size;
	out = (StackOutput *) malloc (size);
	if (out == NULL) {
		register_error (1, "stack_create_output: malloc(StackOutput,%ld) error.\n", size);
		return (NULL);
	}
	memset (out, 0, size);
	out->stack = st;
	strcpy (out->refsta_seed, st->refsta_seed);
	strcpy (out->refsta, st->refsta);
	out->latr = st->latr;
	out->lonr = st->lonr;
	out->channels_tbl = st->channels_tbl;
	strcpy (out->moveout_class_name, st->moveout_class_name);
	out->moveout_class = st->moveout_class;
	out->moveout_params = st->moveout_params;
	strcpy (out->output_variable, output_variable);
	out->pktoutput_channels = pktoutput_channels;
	out->samprate = st->samprate;
	out->nnodes = st->nnodes;
	out->nsamps = nsamps;
	out->gap_value = gdata->gap_value;
	out->clip_value = gdata->clip_value;
	size = (out->nnodes)*(out->nsamps)*sizeof(float);
	out->data = (float *) malloc (size);
	size_out += size;
	if (out->data == NULL) {
		register_error (1, "stack_create_output: malloc(out->data,%ld) error.\n", size);
		free (out);
		return (NULL);
	}
	memset (out->data, 0, size);
	out->tstart = st->buftime - st->buftime_shift;

	if (!strcmp(output_variable, "stack")) {
		for (i=0,k=0; i<out->nsamps; i++) {
			for (j=0; j<out->nnodes; j++,k++) {
				StackNode *node;

				node = (StackNode *) gettbl (st->nodes_tbl, j);
				out->data[k] = node->stack[i];
			}
		}
	} else if (!strcmp(output_variable, "stackpower")) {
		for (i=0,k=0; i<out->nsamps; i++) {
			for (j=0; j<out->nnodes; j++,k++) {
				StackNode *node;

				node = (StackNode *) gettbl (st->nodes_tbl, j);
				out->data[k] = node->stackpower[i];
			}
		}
	} else if (!strcmp(output_variable, "powerstack")) {
		for (i=0,k=0; i<out->nsamps; i++) {
			for (j=0; j<out->nnodes; j++,k++) {
				StackNode *node;

				node = (StackNode *) gettbl (st->nodes_tbl, j);
				out->data[k] = node->powerstack[i];
			}
		}
	} else if (!strcmp(output_variable, "semblance")) {
		for (i=0,k=0; i<out->nsamps; i++) {
			for (j=0; j<out->nnodes; j++,k++) {
				StackNode *node;

				node = (StackNode *) gettbl (st->nodes_tbl, j);
				out->data[k] = node->stackpower[i];
				if (out->data[k] == out->gap_value) continue;
				if (out->data[k] == out->clip_value) continue;
				if (node->powerstack[i] == out->gap_value) {
					out->data[k] = out->gap_value;
					continue;
				}
				if (node->powerstack[i] == out->clip_value) {
					out->data[k] = out->clip_value;
					continue;
				}
				if (node->powerstack[i] == 0.0) {
					if (node->stackpower[i] == 0.0) {
						out->data[k] = (float) 0.0;
					} else {
						out->data[k] = out->gap_value;
					}
					continue;
				}
				out->data[k] /= node->powerstack[i];
			}
		}
	}

	data = wfproc_data_new ("stack");
	if (data == NULL) {
		register_error (0, "stack_create_output: wfproc_data_new() error.\n");
		free (out->data);
		free (out);
		return (NULL);
	}

	data->data = out;
	data->printproc = stack_data_print; 
	data->freeproc = stack_data_free;
	data->copyproc = stack_data_copy;
	data->topktproc = stack_data_topkt;
	data->size = sizeof(WFProcData) + size_out;

	return (data);
}

char *
stack_data_print (WFProcData *data, int verbose)

{
	char line[256];
	char *s;
	StackOutput *out;
	int i, n;
	float min, max;

	out = (StackOutput *) data->data;

	n = out->nsamps*out->nnodes;
	max = out->gap_value;
	min = out->gap_value;
	for (i=0; i<n; i++) {
		if (out->data[i] == out->gap_value) continue;
		if (out->data[i] == out->clip_value) {
			max = out->clip_value;
			min = out->clip_value;
			break;
		}
		if (max == out->gap_value) max = out->data[i];
		else if (out->data[i] > max) max = out->data[i];
		if (min == out->gap_value) min = out->data[i];
		else if (out->data[i] < min) min = out->data[i];
	}

	s = strtime(out->tstart);

	if (max == out->gap_value) {
		sprintf (line, "%s class=%s variable=%s min=gap max=gap", s, out->moveout_class_name, out->output_variable);
	} else {
		sprintf (line, "%s class=%s variable=%s min=%.4f max=%.4f", s, out->moveout_class_name, out->output_variable, min, max);
	}

	free (s);

	return (strdup(line));
}

void
stack_data_free (WFProcData *data)

{
	StackOutput *out;

	if (data == NULL) return;
	out = (StackOutput *) data->data;
	if (out) {
		if (out->data) free (out->data);
		free (out);
	}
	free (data);
}

WFProcData *
stack_data_copy (WFProcData *data)

{
	WFProcData *copy=NULL;
	StackOutput *out, *copyout;
	size_t size;

	copy = wfproc_data_new ("stack");
	if (copy == NULL) {
		register_error (0, "stack_data_copy: wfproc_data_new() error.\n");
		return (NULL);
	}

	*copy = *data;

	size = sizeof(StackOutput);
	copyout = (StackOutput *) malloc(size);
	if (copyout == NULL) {
		register_error (1, "stack_data_copy: malloc(StackOutput,%ld) error.\n", size);
		free (copy);
		return (NULL);
	}

	out = (StackOutput *) data->data;
	*copyout = *out;
	copy->data = copyout;

	size = sizeof(float)*out->nsamps*out->nnodes;
	copyout->data = (float *) malloc (size);
	if (copyout->data == NULL) {
		register_error (1, "stack_data_copy: malloc(copyout->data,%ld) error.\n", size);
		free (copyout);
		free (copy);
		return (NULL);
	}
	memcpy (copyout->data, out->data, size);

	return (copy);
}

int
stack_data_topkt (WFProcData *input, Tbl *translations, Tbl *pkts_tbl)

{
	StackOutput *data;
	int i;
	double calib;
	char segtype[32];
	Packet *pkt;

	data = (StackOutput *) input->data;

	if (data->pktoutput_channels == NULL) return (0);
	if (maxtbl(data->pktoutput_channels) == 0) return (0);

	pkt = newPkt();
	if (pkt == NULL) {
		register_error (0, "stack_data_topkt: newPkt() error.\n");
		return (-1);
	}

	pkt->nchannels = 0;

	calib = 0.001;
	strcpy (segtype, "V");
	if (!strcmp(data->output_variable, "semblance")) {
		calib = 0.001;
		strcpy (segtype, "N");
	} else if (!strcmp(data->output_variable, "powerstack")) {
		calib = 1000.0;
		strcpy (segtype, "Y");
	} else if (!strcmp(data->output_variable, "stackpower")) {
		calib = 1000.0;
		strcpy (segtype, "Y");
	}

	for (i=0; i<maxtbl(data->pktoutput_channels); i++) {
		StackPktoutputType *pktoutput_channel;
		size_t size;
		int j, ret;
		char tnet[32], tsta[32], tchan[32], tloc[32];
		PktChannel *achan, achan_tmp;
		int inode;
	
		pktoutput_channel = (StackPktoutputType *) gettbl (data->pktoutput_channels, i);

		if (strcmp(pktoutput_channel->output_type, "wf")) continue;

		wfproc_split_nscl (pktoutput_channel->seed_name, tnet, tsta, tchan, tloc);

		memset (&achan_tmp, 0, sizeof(PktChannel));
		strcpy (achan_tmp.net, tnet);
		strcpy (achan_tmp.sta, tsta);
		strcpy (achan_tmp.chan, tchan);
		strcpy (achan_tmp.loc, tloc);

		if (translations) {
			ret = wfproc_translate_seedcodes (achan_tmp.net, achan_tmp.sta, achan_tmp.chan, achan_tmp.loc,
							translations, tnet, tsta, tchan, tloc);
			if (ret < 0) {
				register_error (0, "stack_data_topkt: wfproc_translate_seedcodes() error.\n");
				freePkt (pkt);
				return (-1);
			}
			if (ret == 0) {
				continue;
			}
			strcpy (achan_tmp.net, tnet);
			strcpy (achan_tmp.sta, tsta);
			strcpy (achan_tmp.chan, tchan);
			strcpy (achan_tmp.loc, tloc);
		}

		strcpy (achan_tmp.segtype, segtype);
		achan_tmp.calib = calib;
		achan_tmp.calper = -1.0;
		achan_tmp.time = data->tstart;
		achan_tmp.nsamp = data->nsamps;
		achan_tmp.datasz = data->nsamps;
		achan_tmp.samprate = data->samprate;
	
		size = achan_tmp.datasz*sizeof(int);
		achan_tmp.data = (int *) malloc(size);
		if (achan_tmp.data == NULL) {
			register_error (1, "stack_data_topkt: malloc(achan->data,%ld) error.\n", size);
			freePkt (pkt);
			return (-1);
		}

		if (!strcmp(pktoutput_channel->nodes, "max")) {
			for (j=0; j<achan_tmp.nsamp; j++) {
				float fvalue, mvalue;

				mvalue = data->gap_value;
				for (inode=0; inode<data->nnodes; inode++) {
					fvalue = data->data[inode+j*data->nnodes];
					if (fvalue == data->gap_value) {
						continue;
					}
					if (fvalue == data->clip_value) {
						continue;
					}
					if (achan_tmp.calib != 0.0) {
						fvalue /= (float)achan_tmp.calib;
					}
					if (mvalue == data->gap_value) {
						mvalue = fvalue;
					} else if (fvalue > mvalue) {
						mvalue = fvalue;
					}
				}
				if (mvalue == data->gap_value) {
					achan_tmp.data[j] = BADS4;
				} else {
					achan_tmp.data[j] = INT(mvalue);
				}
			}
		} else if (!strcmp(pktoutput_channel->nodes, "max:inode")) {
			achan_tmp.calib = 1.0;
			strcpy (achan_tmp.segtype, "c");
			for (j=0; j<achan_tmp.nsamp; j++) {
				float fvalue, mvalue;
				int ivalue=0;

				achan_tmp.calib = 1.0 ;
				strcpy (achan_tmp.segtype, "c");
				mvalue = data->gap_value;
				for (inode=0; inode<data->nnodes; inode++) {
					fvalue = data->data[inode+j*data->nnodes];
					if (fvalue == data->gap_value) {
						continue;
					}
					if (fvalue == data->clip_value) {
						continue;
					}
					if (mvalue == data->gap_value) {
						mvalue = fvalue;
						ivalue = inode;
					} else if (fvalue > mvalue) {
						mvalue = fvalue;
						ivalue = inode;
					}
				}
				if (mvalue == data->gap_value) {
					achan_tmp.data[j] = BADS4;
				} else {
					achan_tmp.data[j] = ivalue;
				}
			}
		} else if (!strncmp(pktoutput_channel->nodes, "max:", 4)) {
			ret = (*(data->moveout_class->_data_topkt_check)) (data->moveout_params,
						pktoutput_channel->nodes, &achan_tmp);
			if (ret < 0) {
				register_error (0, "stack_data_topkt: %s_data_topkt_check() error.\n",
								data->moveout_class_name);
				freePkt (pkt);
				return (-1);
			}
			if (ret == 0) continue;
			for (j=0; j<achan_tmp.nsamp; j++) {
				float fvalue, mvalue;
				int ivalue=0;

				mvalue = data->gap_value;
				for (inode=0; inode<data->nnodes; inode++) {
					fvalue = data->data[inode+j*data->nnodes];
					if (fvalue == data->gap_value) {
						continue;
					}
					if (fvalue == data->clip_value) {
						continue;
					}
					if (achan_tmp.calib != 0.0) {
						fvalue /= (float)achan_tmp.calib;
					}
					if (mvalue == data->gap_value) {
						mvalue = fvalue;
						ivalue = inode;
					} else if (fvalue > mvalue) {
						mvalue = fvalue;
						ivalue = inode;
					}
				}
				if (mvalue == data->gap_value) {
					achan_tmp.data[j] = BADS4;
				} else {
					ret = (*(data->moveout_class->_data_topkt_data)) (data->moveout_params,
							pktoutput_channel->nodes, ivalue, &(achan_tmp.data[j]));
					if (ret < 0) {
						register_error (0, "stack_data_topkt: %s_data_topkt_data() error.\n",
										data->moveout_class_name);
						freePkt (pkt);
						return (-1);
					}
				}
			}
		} else {
			inode = atoi(pktoutput_channel->nodes);
			if (inode < 0) continue;
			if (inode >= data->nnodes) continue;
			for (j=0; j<achan_tmp.nsamp; j++) {
				float fvalue;

				fvalue = data->data[inode+j*data->nnodes];
				if (fvalue == data->gap_value) {
					achan_tmp.data[j] = BADS4;
					continue;
				}
				if (fvalue == data->clip_value) {
					achan_tmp.data[j] = BADS4;
					continue;
				}
				if (achan_tmp.calib != 0.0) {
					fvalue /= (float)achan_tmp.calib;
				}
				achan_tmp.data[j] = INT(fvalue);
			}
		}
		achan = (PktChannel *) gettbl (pkt->channels, pkt->nchannels);
		if (wfproc_copyPktChannel(&achan, &achan_tmp) < 0) {
			register_error (0, "stack_data_topkt: wfproc_copyPktChannel() error.\n");
			freePkt (pkt);
			return (-1);
		}
		free (achan_tmp.data);
		settbl (pkt->channels, pkt->nchannels, achan);
		(pkt->nchannels)++;
	}

	for (i=pkt->nchannels; i<maxtbl(pkt->channels); i++) {
		PktChannel *achan;

		achan = (PktChannel *) deltbl (pkt->channels, i);
		if (achan) freePktChannel (achan);
	}

	if (pkt->nchannels != 0) {
		if (pkt->nchannels == 1) {
			pkt->pkttype = suffix2pkttype("GENC");
		} else {
			pkt->pkttype = suffix2pkttype("MGENC");
		}
		settbl (pkts_tbl, -1, pkt);
	} else {
		freePkt (pkt);
	}

	for (i=0; i<maxtbl(data->pktoutput_channels); i++) {
		StackPktoutputType *pktoutput_channel;
		BUGrid *grid;
		double dt;
		int ret;
		char tnet[32], tsta[32], tchan[32], tloc[32];

		pktoutput_channel = (StackPktoutputType *) gettbl (data->pktoutput_channels, i);

		if (strcmp(pktoutput_channel->output_type, "grid")) continue;

		wfproc_split_nscl (pktoutput_channel->seed_name, tnet, tsta, tchan, tloc);

		if (translations) {
			char net[32], sta[32], chan[32], loc[32];

			wfproc_split_nscl (pktoutput_channel->seed_name, net, sta, chan, loc);
			ret = wfproc_translate_seedcodes (net, sta, chan, loc,
							translations, tnet, tsta, tchan, tloc);
			if (ret < 0) {
				register_error (0, "stack_data_topkt: wfproc_translate_seedcodes() error.\n");
				return (-1);
			}
			if (ret == 0) {
				continue;
			}
		} else {
			wfproc_split_nscl (pktoutput_channel->seed_name, tnet, tsta, tchan, tloc);
		}
	
		pkt = newPkt();
		if (pkt == NULL) {
			register_error (0, "stack_data_topkt: newPkt() error.\n");
			return (-1);
		}
		pkt->pkttype = suffix2pkttype("gr");
		pkt->pkthook = new_hook (stack_mybugridfree);
		if (pkt->pkthook == NULL) {
			register_error (0, "stack_data_topkt: new_hook() error.\n");
			freePkt (pkt);
			return (-1);
		}
		ret = (*(data->moveout_class->_data_nodes_togrid)) (data->moveout_params, &grid);
		if (ret < 0) {
			register_error (0, "stack_data_topkt: %s_data_nodes_togrid() error.\n", data->moveout_class_name);
			freePkt (pkt);
			return (-1);
		}
		grid->gap_value = data->gap_value;
		grid->data_type = strdup(data->output_variable);
		grid->params[0] = data->latr;
		grid->params[1] = data->lonr;
		if (data->samprate == 0.0) {
			dt = 0.0;
		} else {
			dt = 1.0/data->samprate;
		}
		if (bugrid_putanddecimate_bugrid (grid, data->tstart, data->nsamps, dt, data->data, data->stack->grid_decimation) < 0) {
			register_error (0, "stack_data_topkt: bugrid_putanddecimate_bugrid() error.\n");
			bugrid_free_bugrid (grid);
			freePkt (pkt);
			return (-1);
		}
		pkt->pkthook->p = grid;
		memcpy (&(pkt->pkthook->type), "grid", 4);
		strcpy (pkt->parts.src_net, tnet);
		strcpy (pkt->parts.src_sta, tsta);
		strcpy (pkt->parts.src_chan, tchan);
		strcpy (pkt->parts.src_loc, tloc);
		strcpy (pkt->parts.src_suffix, "gr");
		settbl (pkts_tbl, -1, pkt);
	}

	return (0);
}

static void
stack_mybugridfree (void *ptr)

{
	BUGrid *grid = (BUGrid *) ptr;

	bugrid_free_bugrid(grid);
}
