/* Copyright (c) 2008 Boulder Real Time Technologies, Inc. */
/* All rights reserved */
 
/* This software module is wholly owned by Boulder Real Time 
   Technologies, Inc. This software may be used freely in any 
   way as long as the copyright statement above is not removed. */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <strings.h>

#include "wfstats.h"
#include "Pkt.h"

void *
wfstats_new (Pf *params)

{
	size_t size;
	Wfstats *wfst;

	size = sizeof(Wfstats);
	wfst = (Wfstats *) malloc (size);
	if (wfst == NULL) {
		register_error (1, "wfstats_new: malloc(Wfstats,%ld) error.\n", size);
		return (NULL);
	}

	memset (wfst, 0, size);

	return (wfst);
}

int
wfstats_destroy (void *wfstv)

{
	Wfstats *wfst = (Wfstats *) wfstv;

	if (wfst == NULL) return (0);

	free (wfst);

	return (0);
}

int
wfstats_config (void *wfstv, Pf *params, Pf *state)

{
	Wfstats *wfst = (Wfstats *) wfstv;

	if (parse_param (params, "verbose", P_LINT, 0, &(wfst->verbose)) < 0) {
		register_error (0, "wfstats_config: parse_param(verbose) error.\n");
		return (-1);
	}
	if (parse_param (params, "accept", P_TBL, 0, &(wfst->accept_tbl)) < 0) {
		register_error (0, "wfstats_config: parse_param(accept) error.\n");
		return (-1);
	}
	if (parse_param (params, "reject", P_TBL, 0, &(wfst->reject_tbl)) < 0) {
		register_error (0, "wfstats_config: parse_param(reject) error.\n");
		return (-1);
	}

	return (0);
}

int
wfstats_proc (void *wfstv, WFProcData *input, Tbl **output, Pf **state)

{
	Wfstats *wfst = (Wfstats *) wfstv;
	WFProcTask *this_task;
	PktChannel *data;
	char nscl[128];
	int i, nbad, first;
	float datamin, datamax;
	double mean, std;
	WFProcData *outdataobj;
	WfstatsOutput *out;
	size_t size;

	this_task = wfproc_task_get_from_obj (wfstv);

	*output = newtbl(0);
	if (*output == NULL) {
		register_error (0, "wfstats_proc: newtbl(output) error.\n");
		wfproc_data_free (input);
		return (-1);
	}

	if (strcmp(input->type, "wf")) {
		if (wfst->verbose) {
			elog_notify (0, "%s: inputtask=%s: inputtype=%s: Skipping because of wrong type\n", 
						this_task->name, input->task->name, input->type);
		}
		wfproc_data_free (input);
		return (0);
	}

	data = (PktChannel *) input->data;

	wfproc_make_nscl (data->net, data->sta, data->chan, data->loc, nscl);

	if (wfst->accept_tbl) {
		Tbl *accept;

		accept = tblgrep (wfst->accept_tbl, nscl);
		if (accept == NULL || maxtbl(accept) < 1) {
			if (wfst->verbose) {
				elog_notify (0, "%s: inputtask=%s: Skipping because %s doesn't match accept_tbl\n", 
						this_task->name, input->task->name, nscl);
			}
			wfproc_data_free (input);
			if (accept) freetbl (accept, NULL);
			return (0);
		}
		if (accept) freetbl (accept, NULL);
	}

	if (wfst->reject_tbl) {
		Tbl *reject;

		reject = tblgrep (wfst->reject_tbl, nscl);
		if (reject != NULL && maxtbl(reject) > 0) {
			if (wfst->verbose) {
				elog_notify (0, "%s: inputtask=%s: Skipping because %s does match reject_tbl\n", 
						this_task->name, input->task->name, nscl);
			}
			freetbl (reject, NULL);
			wfproc_data_free (input);
			return (0);
		}
		if (reject) freetbl (reject, NULL);
	}

	nbad = 0;
	datamin = 0;
	datamax = 0;
	first = 1;
	mean = 0.0;
	std = 0.0;
	for (i=0; i<data->nsamp; i++) {
		float fdata;

		if (data->isfloat) {
			memcpy (&fdata, &(data->data[i]), sizeof(float));
			if (fdata == wfproc_float_gap_value) {
				nbad++;
				continue;
			}
		} else {
			if (data->data[i] == BADS4) {
				nbad++;
				continue;
			}
			fdata = data->data[i];
		}
		if (first) {
			datamin = fdata;
			datamax = fdata;
			first = 0;
		} else if (fdata > datamax) {
			datamax = fdata;
		} else if (fdata < datamin) {
			datamin = fdata;
		}
		mean += fdata;
		std += fdata*fdata;
	}
	if (nbad < data->nsamp) {
		mean /= (data->nsamp-nbad);
		std /= (data->nsamp-nbad);
		std -= mean*mean;
		if (std > 0.0) std = sqrt(std);
	}

	if (wfst->verbose) {
		char *s;

		s = epoch2str (data->time, "%Y%j:%H:%M:%S.%s");
		elog_notify (0, "%s: %s: %s: nsamp=%d nbad=%d min=%.3f max=%.3f mean=%.3f std=%.3f\n", 
					this_task->name, s, nscl, data->nsamp, nbad, 
					datamin, datamax, mean, std);
		free (s);
	}

	size = sizeof(WfstatsOutput);
	out = (WfstatsOutput *) malloc (size);
	if (out == NULL) {
		register_error (1, "wfstats_proc: malloc(WfstatsOutput,%ld) error.\n", size);
		wfproc_data_free (input);
		return (0);
	}
	memset (out, 0, size);
	
	strcpy (out->net, data->net);
	strcpy (out->sta, data->sta);
	strcpy (out->chan, data->chan);
	strcpy (out->loc, data->loc);
	out->time = data->time;
	out->samprate = data->samprate;
	out->nsamp = data->nsamp;
	out->nbad = nbad;
	out->min = datamin;
	out->max = datamax;
	out->mean = mean;
	out->std = std;

	outdataobj = wfproc_data_new ("wfstats");
	if (outdataobj == NULL) {
		register_error (0, "wfstats_proc: wfproc_data_new() error.\n");
		free (out);
		wfproc_data_free (input);
		return (0);
	}

	outdataobj->data = out;
	outdataobj->copyproc = wfstats_data_copy;
	outdataobj->freeproc = wfstats_data_free;
	outdataobj->topktproc = wfstats_data_topkt;
	outdataobj->printproc = wfstats_data_print; 
	outdataobj->size = sizeof(WFProcData) + sizeof(WfstatsOutput);

	settbl (*output, -1, outdataobj);

	wfproc_data_free (input);

	return (0);
}

WFProcData *
wfstats_data_copy (WFProcData *data)

{
	WFProcData *copy=NULL;
	WfstatsOutput *out, *copyout;
	size_t size;

	copy = wfproc_data_new ("wfstats");
	if (copy == NULL) {
		register_error (0, "wfstats_data_copy: wfproc_data_new() error.\n");
		return (NULL);
	}

	*copy = *data;

	size = sizeof(WfstatsOutput);
	copyout = (WfstatsOutput *) malloc(size);
	if (copyout == NULL) {
		register_error (1, "wfstats_data_copy: malloc(WfstatsOutput,%ld) error.\n", size);
		free (copy);
		return (NULL);
	}

	out = (WfstatsOutput *) data->data;
	*copyout = *out;
	copy->data = copyout;

	return (copy);
}

void 
wfstats_data_free (WFProcData *data)

{
	WfstatsOutput *out;

	if (data == NULL) return;
	out = (WfstatsOutput *) data->data;
	if (out) free (out);

	free (data);
}

int 
wfstats_data_topkt (WFProcData *data, Tbl *translations, Tbl *pkts_tbl)

{
	WfstatsOutput *wfstatsdata;
	int ret;
	Packet *pkt;
	char tnet[32], tsta[32], tchan[32], tloc[32];

	wfstatsdata = (WfstatsOutput *) data->data;
	pkt = newPkt();
	if (pkt == NULL) {
		register_error (0, "wfstats_data_topkt: newPkt() error.\n");
		return (-1);
	}
	pkt->pkttype = suffix2pkttype("log");

	if (translations) {
		ret = wfproc_translate_seedcodes (wfstatsdata->net, wfstatsdata->sta, wfstatsdata->chan, wfstatsdata->loc,
							translations, tnet, tsta, tchan, tloc);
		if (ret < 0) {
			register_error (0, "wfstats_data_topkt: wfproc_translate_seedcodes() error.\n");
			freePkt (pkt);
			return (-1);
		}
		if (ret == 0) {
			freePkt (pkt);
			return (0);
		}
	} else {
		strcpy (tnet, wfstatsdata->net);
		strcpy (tsta, wfstatsdata->sta);
		strcpy (tchan, wfstatsdata->chan);
		strcpy (tloc, wfstatsdata->loc);
	}

	strcpy (pkt->parts.src_net, tnet);
	strcpy (pkt->parts.src_sta, tsta);
	strcpy (pkt->parts.src_chan, tchan);
	strcpy (pkt->parts.src_loc, tloc);
	pkt->time = wfstatsdata->time;
	pkt->string = wfstats_data_print (data, 1);
	if (pkt->string == NULL) {
		register_error (0, "wfstats_data_topkt: wfstats_data_print() error.\n");
		freePkt (pkt);
		return (-1);
	}
	pkt->string_size = strlen(pkt->string);

	settbl (pkts_tbl, -1, pkt);

	return (0);
}

char *
wfstats_data_print (WFProcData *data, int verbose)

{
	WfstatsOutput *wfstatsdata;
	char line[512];
	char *msg=NULL;
	char *s;
	char nscl[128];

	if (verbose == 0) return (NULL);

	wfstatsdata = (WfstatsOutput *) data->data;
	s = epoch2str (wfstatsdata->time, "%Y%j:%H:%M:%S.%s");
	wfproc_make_nscl (wfstatsdata->net, wfstatsdata->sta, wfstatsdata->chan, wfstatsdata->loc, nscl);
	sprintf (line, "%s: %s: nsamp=%d nbad=%d min=%.3f max=%.3f mean=%.3f std=%.3f", 
			s, nscl, wfstatsdata->nsamp, wfstatsdata->nbad, wfstatsdata->min, 
			wfstatsdata->max, wfstatsdata->mean, wfstatsdata->std);
	free (s);
	msg = strdup(line);

	return (msg);
}
